/* Copyright (C) 1997, 1999, 2001, 2002, 2003, 2005, 2006
   Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Thorsten Kukuk <kukuk@vt.uni-paderborn.de>, 1997.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <atomic.h>
#include <nss.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <bits/libc-lock.h>
#include <rpcsvc/nis.h>

#include "nss-nisplus.h"
#include "nisplus-parser.h"
#include <libnsl.h>


__libc_lock_define_initialized (static, lock)

/* Previous result of iteration.  */
static nis_result *result;

/* All results of batch table load.  */
static nis_result *cached_results;
static size_t cached_results_iter;

nis_name pwd_tablename_val attribute_hidden;
size_t pwd_tablename_len attribute_hidden;

enum nss_status
_nss_pwd_create_tablename (int *errnop)
{
  if (pwd_tablename_val == NULL)
    {
      const char *local_dir = nis_local_directory ();
      size_t local_dir_len = strlen (local_dir);
      static const char prefix[] = "passwd.org_dir.";

      char *p = malloc (sizeof (prefix) + local_dir_len);
      if (p == NULL)
	{
	  *errnop = errno;
	  return NSS_STATUS_TRYAGAIN;
	}

      memcpy (__stpcpy (p, prefix), local_dir, local_dir_len + 1);

      pwd_tablename_len = sizeof (prefix) - 1 + local_dir_len;

      atomic_write_barrier ();

      if (atomic_compare_and_exchange_bool_acq (&pwd_tablename_val, p, NULL))
	{
	  /* Another thread already installed the value.  */
	  free (p);
	  pwd_tablename_len = strlen (pwd_tablename_val);
	}
    }

  return NSS_STATUS_SUCCESS;
}


static void
internal_nisplus_endpwent (void)
{
  if (cached_results != NULL)
    {
      nis_freeresult (cached_results);
      cached_results = NULL;
      cached_results_iter = 0;
    }

  if (result != NULL)
    {
      nis_freeresult (result);
      result = NULL;
    }
}


static enum nss_status
internal_nisplus_setpwent (int *errnop)
{
  enum nss_status status;

  cached_results = nis_list (pwd_tablename_val, FOLLOW_PATH | FOLLOW_LINKS,
			     NULL, NULL);

  if (cached_results == NULL)
    {
      *errnop = errno;
      status = NSS_STATUS_TRYAGAIN;
    }
  else if (__builtin_expect ((status = niserr2nss (cached_results->status))
			     != NSS_STATUS_SUCCESS, 0))
    {
      nis_freeresult (cached_results);
      cached_results = NULL;
    }
  else if (__builtin_expect (__type_of (NIS_RES_OBJECT (cached_results))
			     != NIS_ENTRY_OBJ
			     || strcmp (NIS_RES_OBJECT (cached_results)->EN_data.en_type,
					"passwd_tbl") != 0
			     || NIS_RES_OBJECT (cached_results)->EN_data.en_cols.en_cols_len < 7,
			     0))
    {
      nis_freeresult (cached_results);
      cached_results = NULL;
      status = NSS_STATUS_NOTFOUND;
    }

  return status;
}


enum nss_status
_nss_nisplus_setpwent (int stayopen)
{
  enum nss_status status = NSS_STATUS_SUCCESS;

  __libc_lock_lock (lock);

  internal_nisplus_endpwent ();

  if (pwd_tablename_val == NULL)
    {
      // XXX We need to be able to set errno.  Pass in new parameter.
      int err;
      status = _nss_pwd_create_tablename (&err);
    }

  if (status == NSS_STATUS_SUCCESS
      && (_nsl_default_nss () & NSS_FLAG_SETENT_BATCH_READ))
    {
      // XXX We need to be able to set errno.  Pass in new parameter.
      int err;
      status = internal_nisplus_setpwent (&err);
    }

  __libc_lock_unlock (lock);

  return status;
}


enum nss_status
_nss_nisplus_endpwent (void)
{
  __libc_lock_lock (lock);

  internal_nisplus_endpwent ();

  __libc_lock_unlock (lock);

  return NSS_STATUS_SUCCESS;
}


static enum nss_status
internal_nisplus_getpwent_r (struct passwd *pw, char *buffer, size_t buflen,
			     int *errnop)
{
  int parse_res = -1;
  nis_result *saved_res = NULL;

  /* Get the next entry until we found a correct one. */
  do
    {
      if (cached_results != NULL)
	{
	handle_batch_read:
	  /* See whether we reported the last problem.  */
	  if (cached_results_iter >= NIS_RES_NUMOBJ (cached_results))
	    return NSS_STATUS_NOTFOUND;

	  parse_res = _nss_nisplus_parse_pwent (cached_results,
						cached_results_iter, pw,
						buffer, buflen, errnop);
	}
      else
	{
	  if (result == NULL)
	    {
	      if (pwd_tablename_val == NULL)
		{
		  enum nss_status status = _nss_pwd_create_tablename (errnop);

		  if (status != NSS_STATUS_SUCCESS)
		    return status;
		}

	      /* Determine whether we should instead read all entries at
		 once.  */
	      if (_nsl_default_nss () & NSS_FLAG_SETENT_BATCH_READ)
		{
		  enum nss_status status = internal_nisplus_setpwent (errnop);

		  if (status == NSS_STATUS_SUCCESS && cached_results != NULL)
		    goto handle_batch_read;
		}

	      saved_res = NULL;

	      result = nis_first_entry (pwd_tablename_val);
	      if (result == NULL)
		{
		  *errnop = errno;
		  return NSS_STATUS_TRYAGAIN;
		}
	      if (niserr2nss (result->status) != NSS_STATUS_SUCCESS)
		return niserr2nss (result->status);
	    }
	  else
	    {
	      saved_res = result;
	      result = nis_next_entry (pwd_tablename_val, &result->cookie);
	      if (result == NULL)
		{
		  *errnop = errno;
		  return NSS_STATUS_TRYAGAIN;
		}
	      if (niserr2nss (result->status) != NSS_STATUS_SUCCESS)
		{
		  nis_freeresult (saved_res);
		  return niserr2nss (result->status);
		}
	    }

	  parse_res = _nss_nisplus_parse_pwent_chk (result, pw, buffer,
						    buflen, errnop);
	}

      if (__builtin_expect (parse_res == -1, 0))
	{
	  if (cached_results == NULL)
	    {
	      nis_freeresult (result);
	      result = saved_res;
	    }
	  *errnop = ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	}

      if (cached_results != NULL)
	++cached_results_iter;
      else
	if (saved_res)
	  {
	    nis_freeresult (saved_res);
	    saved_res = NULL;
	  }
    }
  while (!parse_res);

  return NSS_STATUS_SUCCESS;
}

enum nss_status
_nss_nisplus_getpwent_r (struct passwd *result, char *buffer, size_t buflen,
			 int *errnop)
{
  int status;

  __libc_lock_lock (lock);

  status = internal_nisplus_getpwent_r (result, buffer, buflen, errnop);

  __libc_lock_unlock (lock);

  return status;
}

enum nss_status
_nss_nisplus_getpwnam_r (const char *name, struct passwd *pw,
			 char *buffer, size_t buflen, int *errnop)
{
  int parse_res;

  if (pwd_tablename_val == NULL)
    {
      enum nss_status status = _nss_pwd_create_tablename (errnop);

      if (status != NSS_STATUS_SUCCESS)
	return status;
    }

  if (name == NULL)
    {
      *errnop = EINVAL;
      return NSS_STATUS_UNAVAIL;
    }

  nis_result *result;
  char buf[strlen (name) + 9 + pwd_tablename_len];
  int olderr = errno;

  snprintf (buf, sizeof (buf), "[name=%s],%s", name, pwd_tablename_val);

  result = nis_list (buf, FOLLOW_PATH | FOLLOW_LINKS, NULL, NULL);

  if (result == NULL)
    {
      *errnop = ENOMEM;
      return NSS_STATUS_TRYAGAIN;
    }

  if (__builtin_expect (niserr2nss (result->status) != NSS_STATUS_SUCCESS, 0))
    {
      enum nss_status status =  niserr2nss (result->status);

      __set_errno (olderr);

      nis_freeresult (result);
      return status;
    }

  parse_res = _nss_nisplus_parse_pwent_chk (result, pw, buffer, buflen,
					    errnop);

  nis_freeresult (result);

  if (__builtin_expect (parse_res < 1, 0))
    {
      if (parse_res == -1)
	{
	  *errnop = ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	}
      else
	{
	  __set_errno (olderr);
	  return NSS_STATUS_NOTFOUND;
	}
    }

  return NSS_STATUS_SUCCESS;
}

enum nss_status
_nss_nisplus_getpwuid_r (const uid_t uid, struct passwd *pw,
			 char *buffer, size_t buflen, int *errnop)
{
  if (pwd_tablename_val == NULL)
    {
      enum nss_status status = _nss_pwd_create_tablename (errnop);

      if (status != NSS_STATUS_SUCCESS)
	return status;
    }

  int parse_res;
  nis_result *result;
  char buf[8 + 3 * sizeof (unsigned long int) + pwd_tablename_len];
  int olderr = errno;

  snprintf (buf, sizeof (buf), "[uid=%lu],%s",
	    (unsigned long int) uid, pwd_tablename_val);

  result = nis_list (buf, FOLLOW_PATH | FOLLOW_LINKS, NULL, NULL);

  if (result == NULL)
    {
      *errnop = ENOMEM;
      return NSS_STATUS_TRYAGAIN;
    }

  if (__builtin_expect (niserr2nss (result->status) != NSS_STATUS_SUCCESS, 0))
    {
      enum nss_status status = niserr2nss (result->status);

      __set_errno (olderr);

      nis_freeresult (result);
      return status;
    }

  parse_res = _nss_nisplus_parse_pwent_chk (result, pw, buffer, buflen,
					    errnop);

  nis_freeresult (result);

  if (__builtin_expect (parse_res < 1, 0))
    {
      if (parse_res == -1)
	{
	  *errnop = ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	}
      else
	{
	  __set_errno (olderr);
	  return NSS_STATUS_NOTFOUND;
	}
    }

  return NSS_STATUS_SUCCESS;
}
